from pathlib import Path
import importlib
import numpy as np
import sys
this_dir = Path(__file__).parent
sys.path.insert(0, str(this_dir.parent))
from mri_distortion_toolkit.PhantomDesigner import PhantomSlice

"""
This script was set up to demonstrate some of the different phantom designs that can be generated with this code
"""
different_designs =['elipsoid_cartesian', 'rectangle_polar', 'circle_spiral']
design_to_build = different_designs[1]  # change this to change slice design
Nslices = 1  # just need one slice for demo
SliceZPositions = 0

def test_elipsoid_cartesian():
    Slice = PhantomSlice(slice_shape='ellipse',
                                           slice_thickness=35, HVL_x=200, HVL_Y=160,
                                           hole_depth=17, hole_spacing=20,
                                           hole_radius=8.7 / 2,
                                           z_pos=-SliceZPositions,
                                           HoleCentroids='cartesian',
                                           DSV=None,
                                           bottom_cut=50)


def test_rectangle_polar():
    Slice = PhantomSlice(slice_shape='rectangle',
                                           slice_thickness=20, HVL_x=100, HVL_Y=100,
                                           hole_depth=17, hole_spacing=20,
                                           hole_radius=5,
                                           z_pos=-SliceZPositions,
                                           HoleCentroids='ROI_polar',
                                           DSV=50,
                                           bottom_cut=0)


def test_circle_spiral():
    # lets generate a spiral of markers just for fun
    theta = np.radians(np.linspace(0, 360*5, 100))
    r = theta ** 2
    r = (r / np.max(r)) * 180 # scale to slice dimension
    del_ind = r < 30  # remove overlapping centroids
    r = np.delete(r, del_ind)
    theta = np.delete(theta, del_ind)
    r = np.insert(r, 0, 0)  # put back in zero point
    theta = np.insert(theta, 0, 0)
    x_2 = r * np.cos(theta)
    y_2 = r * np.sin(theta)
    hole_centroids = [x_2, y_2]

    Slice = PhantomSlice(slice_shape='ellipse',
                                           slice_thickness=35, HVL_x=200, HVL_Y=200,
                                           hole_depth=17, hole_spacing=20,
                                           hole_radius=8.7 / 2,
                                           z_pos=-SliceZPositions,
                                           HoleCentroids=hole_centroids,
                                           DSV=None,
                                           bottom_cut=0)




import sys
from pathlib import Path
import numpy as np
import pandas as pd
this_dir = Path(__file__).parent
test_data_dir = this_dir / 'test_data'
sys.path.insert(0, str(this_dir.parent))
from mri_distortion_toolkit.Harmonics import SphericalHarmonicFit
from mri_distortion_toolkit import calculate_harmonics
from mri_distortion_toolkit.MarkerAnalysis import MarkerVolume
from mri_distortion_toolkit.utilities import convert_spherical_harmonics
from mri_distortion_toolkit.utilities import get_dicom_data


def test_spherical_harmonics_stability():
    """
    tests the creation of a SphericalHarmonicFit from an opera table file, and checks that
    the fitted harmonmics closely match a previously created set from the same data
    """

    data_loc = this_dir / 'test_data' / 'OperaTableTest.table'
    data = np.loadtxt(str(data_loc), skiprows=6)
    data_pd =  pd.DataFrame(data, columns = ['x','y','z','Bz'])
    harmonics = SphericalHarmonicFit(data_pd, n_order=10)

    test_harmonics = \
        pd.read_csv((this_dir / 'test_data'/ 'ground_truth_harmonics.csv').resolve(), index_col=0).squeeze("columns")
    assert np.allclose(harmonics.harmonics, test_harmonics, rtol=1e-05, atol=1e-06)


def test_calculate_harmonics():
    """
    test our wrapper function; just checks for stability
    """
    distorted_volume = MarkerVolume(test_data_dir / 'MR.mrk.json')
    ground_truth_volume = MarkerVolume(test_data_dir / 'CT.mrk.json')
    dicom_data_loc = test_data_dir / 'dicom_data.json'
    dicom_data = get_dicom_data(dicom_data_loc)
    gradient_strength = np.array(dicom_data['gradient_strength']) * 1e3

    FieldData = pd.read_csv(test_data_dir / 'MagneticFieldData.csv', index_col=0).squeeze("columns")
    dicom_data_loc = Path(test_data_dir / 'dicom_data.json')  # previosly saved from a MarkerVolume
    dicom_data = get_dicom_data(dicom_data_loc)
    gradient_strength = np.array(dicom_data['gradient_strength']) * 1e3
    normalisation_factor = [1 / gradient_strength[0], 1 / gradient_strength[1], 1 / gradient_strength[2],
                            1]  # this normalised gradient harmonics to 1mT/m

    G_x_Harmonics, G_y_Harmonics, G_z_Harmonics, B0_Harmonics = calculate_harmonics(FieldData,
                                                                                    scale=normalisation_factor,
                                                                                    n_order=8)

    # can't be bothered saving in and reading all harmonics so just checking max:
    assert np.allclose(G_x_Harmonics.harmonics.max(), .40997688028675766)
    assert np.allclose(G_y_Harmonics.harmonics.max(), .7000613996928333)
    assert np.allclose(G_z_Harmonics.harmonics.max(), .5690328634698217)


def test_harmonic_conversion():
    """
    test that we can read in harmonics, convert them forward and back and the remain the same
    """
    g_x_harmonics = pd.read_csv(test_data_dir / 'G_x_Harmonics.csv', index_col=0).squeeze("columns")
    g_x_harmonics_no_norm = convert_spherical_harmonics(g_x_harmonics, input_format='full', output_format='none')
    g_x_harmonics_norm = convert_spherical_harmonics(g_x_harmonics_no_norm, input_format='none', output_format='full')
    assert np.allclose(g_x_harmonics, g_x_harmonics_norm)
    g_x_harmonics_numpy = g_x_harmonics.to_numpy()
    g_x_harmonics_no_norm_numpy = convert_spherical_harmonics(g_x_harmonics_numpy, input_format='full', output_format='none')
    assert np.allclose(g_x_harmonics_no_norm, g_x_harmonics_no_norm_numpy)


def test_scale_harmonics():
    data_loc = this_dir / 'test_data' / 'OperaTableTest.table'
    data = np.loadtxt(str(data_loc), skiprows=6)
    data_pd =  pd.DataFrame(data, columns = ['x', 'y', 'z', 'Bz'])
    harmonics = SphericalHarmonicFit(data_pd, n_order=10)
    harmonics_scaled = SphericalHarmonicFit(data_pd, n_order=10, scale=10)
    scale_factor = np.mean(np.divide(harmonics_scaled.harmonics, harmonics.harmonics))
    assert np.isclose(10, scale_factor)


import os.path
import sys
from pathlib import Path
import numpy as np
import pandas as pd


this_dir = Path(__file__).parent
sys.path.insert(0, str(this_dir.parent))
from mri_distortion_toolkit.MarkerAnalysis import MarkerVolume, MatchedMarkerVolumes
from mri_distortion_toolkit.utilities import get_dicom_data


def test_mr_dicom_data_read_in():
    """
    - test that we read in dicom data and return a number at least in the vicinity of the right answer
    - tests that the dicom data exists and frequency encoding direction is correctly extracted
    - tests that we can export this data to slicer format, read it back in, and that those two datasets are correctly
        matched
    - tests that we can export and read back in the dicom_data
    """

    volume = MarkerVolume((this_dir / 'test_data' / 'MR_dicom').resolve())
    assert volume.MarkerCentroids.shape[0] > 50
    assert volume.MarkerCentroids.shape[0] < 65
    # there should be 58 I think so this is a very loose test but should catch gross errors

    assert volume.dicom_data is not None
    # test manual cutoff works:
    volume2 = MarkerVolume((this_dir / 'test_data' / 'MR_dicom').resolve(), threshold=volume._cutoffpoint)
    assert volume.MarkerCentroids.shape == volume2.MarkerCentroids.shape
    assert volume.dicom_data['freq_encode_direction'] == 'x'

    volume.export_to_slicer(this_dir / 'test_data', filename='dicom_export_test')
    volume.save_dicom_data(this_dir / 'test_data', filename='dicom_data_export_test')
    new_dicom_data = get_dicom_data(this_dir / 'test_data' / 'dicom_data_export_test.json')
    assert new_dicom_data == volume.dicom_data
    volume2 = MarkerVolume(str(this_dir / 'test_data' / 'dicom_export_test.mrk.json'))
    assert np.allclose(volume.MarkerCentroids, volume2.MarkerCentroids, rtol=1e-01, atol=1e-01)


def test_ct_dicom_data_read_in():
    """
    - test that we read in dicom data and return a number at least in the vicinity of the right answer
    - tests that the dicom data exists and frequency encoding direction is correctly extracted
    - tests that we can export this data to slicer format, read it back in, and that those two datasets are correctly
        matched
    - tests that we can export and read back in the dicom_data
    """

    volume = MarkerVolume((this_dir / 'test_data' / 'CT_dicom').resolve())
    assert volume.MarkerCentroids.shape[0] > 1
    assert volume.MarkerCentroids.shape[0] < 4
    # there should be 3 I think so this is a very loose test but should catch gross errors
    assert volume.dicom_data is None


def test_pandas_read_in():
    """
    test we can read in numpy data and that the marker volume is the correct shape
    """
    r_outer = 150
    test_data = np.random.rand(100, 3) * r_outer
    test_data = pd.DataFrame(test_data, columns=['x', 'y', 'z'])
    volume = MarkerVolume(test_data)
    assert volume.MarkerCentroids.shape[0] == test_data.shape[0]
    volume.perturb_marker_positions(1)  # just so this code gets executed


def test_json_read_in():
    """
    test we can read in json data and that the marker volume is the correct shape
    """
    data_loc = this_dir / 'test_data' / 'CT.mrk.json'
    volume = MarkerVolume(data_loc)
    assert volume.MarkerCentroids.shape[0] == 336


def test_marker_matching_rigid():
    """
    this tests that two volumes offset from each other by a known rigid amount then randomised
    are correctly matched
    so far it always passes, but I think there is potential for it to fail if two markers are
    generated very close together, so we may need to add some extra condition on spacing
    """
    r_outer = 150  # just to scale random numbers with
    data_offset = 2
    expected_distance = np.sqrt(3*(data_offset**2))
    test_data = np.random.rand(100, 3) * r_outer
    test_data2 = test_data + data_offset
    np.random.shuffle(test_data2)  # randomise second dataset
    test_data = pd.DataFrame(test_data, columns=['x', 'y', 'z'])
    test_data2 = pd.DataFrame(test_data2, columns=['x', 'y', 'z'])
    volume1 = MarkerVolume(test_data)
    volume2 = MarkerVolume(test_data2)

    matched_volume = MatchedMarkerVolumes(volume1, volume2, sorting_method='radial')
    assert np.allclose(matched_volume.MatchedCentroids.match_distance, expected_distance)
    matched_volume = MatchedMarkerVolumes(volume1, volume2, sorting_method='nearest')
    assert np.allclose(matched_volume.MatchedCentroids.match_distance, expected_distance)




from pathlib import Path
import sys
import numpy as np

this_file_loc = Path(__file__).parent.resolve()
examples_loc = this_file_loc.parent / 'examples'
sys.path.insert(0, str(examples_loc))

def test_examples_run():
    """
    simply imports all examples: since they are scripts, they execute on import.
    The variables of each example are available under the import name space
    e.g. everything created in _2_marker_matching is under _2_marker_matching.relevant_variable
    """
    from examples import _0_build_phantoms
    from examples import _0b_complete_phantom
    from examples import _2_marker_matching
    from examples import _3_field_calculation
    from examples import _4_fit_harmonics
    from examples import _5_reporting

    assert np.mean(_5_reporting.report._MatchedMarkerVolume.abs_dis) < 10  # was 8.73, could change a bit with changes to other algorithms
    assert np.max(_5_reporting.report._MatchedMarkerVolume.abs_dis) < 100  # was 71

